import logging
from datetime import datetime
from logging.handlers import TimedRotatingFileHandler

from sqlalchemy.orm import Session

from .base import BaseNetworkClient

# from vc.
from vc.client.soldo import user, wallets, card, group, order, transaction
from vc.models.soldo import WalletSo, CardSo
from vc.settings import Settings
from .soldo_event import EventMixer
from vc.libs.utils import set_config

logger = logging.getLogger(__name__)
set_config(logger, filename="soldo2.log")


class SoldoException(Exception):

    def __init__(self, msg: str, response_data: dict):
        self.msg = msg
        self.response_data = response_data

    def __str__(self):
        return f"{self.msg}, {str(self.response_data)}"


class Soldo(EventMixer, BaseNetworkClient):
    settings = Settings({
        "ACCESS_TOKEN": "ArKNudyjPu7HgWj6V8QiaBfBS8xPlsSG",
    })
    event_list = ["new_user", "wallet_created", "store_order_completed"]
    __cache = {'c274b136-5999-4626-850c-46f5db5e5473':
                   {"id": "1d4df4f9-0c89-4913-8d68-8e6c9d19c611", 'wallet_id': 1, 'status': 'PLACED',
                    "category": "CARD"}, }

    def get_cache(self, key: str):
        return self.__cache.get(key)

    def __init__(self, name, uri,
                 api_url: str,
                 client_id: str,
                 client_secret: str,
                 group_id: str,
                 token: str,
                 wallet_safe: str,
                 filepath_private_pem: str, log_file: str = None, currency="USD", user_model=None, **config):
        data = dict(name=name, currency=currency,
                    CLIENT_ID=client_id,
                    CLIENT_SECRET=client_secret,
                    API_URL=api_url,
                    TOKEN=token,
                    WALLET_SAFE=wallet_safe,
                    USER_MODEL=user_model,
                    LOG_FILE=log_file,
                    PATH_RSA_PRIVATE=filepath_private_pem,
                    GROUP_ID=group_id, **config)
        set_config(logger, filename=Soldo.settings.LOG_FILE)
        Soldo.settings.update_config(**data)
        super().__init__(uri, user_model=user_model, **config)

    def activate_service(self, db: Session, user_id: int, owner_type="company"):
        user = db.query(self._user).filter(self._user.id == user_id).first()
        logger.debug(user.__dict__)
        response_data = wallets.create(owner_type, self.settings.currency, name=user.email, custom_reference_id=user.id)
        order = response_data.data
        if not order.is_valid or order.status != "PLACED":
            raise SoldoException("Error create_wallet", response_data.dict())

        wallet = WalletSo(user_id=user.id, currency=self.settings.currency, )
        if order.items:
            wallet.search_id = order.items[0].id

        self.save_obj(db, wallet)
        return wallet

    def get_transactions_by_wallet_id(self, wallet_id, type="wallet", **kwargs):
        wallet_q = db.query(WalletSo).filter(WalletSo.id == wallet_id).first()
        response_data = transaction.search(type=type, publicId=wallet_q.search_id, **kwargs).data
        return response_data

    def get_statements_by_card_id(self, db: Session, card_id, page=0, **kwargs):
        card_q = db.query(CardSo).filter(CardSo.id==card_id).first()
        response_data = transaction.search(type="card", publicId=card_q.search_id,**kwargs).data
        return response_data

    def upload_cards(self, db: Session):
        response_cards = card.search(page_size=1000500, type="company").data.results
        query_cards = db.query(CardSo).filter(CardSo.search_id is not None).all()

        list_model_id = []
        list_model = []
        for w in query_cards:
            list_model_id.append(w.search_id)
            list_model.append({"search_id": w.search_id, "id": w.id})

        filter_model = list(filter(
            lambda w: w.id in list_model_id,
            response_cards
        ))

        update_card = []
        for model in filter_model:
            response_wallet = list(filter(
                lambda w: w.get("search_id") == model.id,
                list_model
            ))
            if response_wallet:
                update_card.append({
                    "id": response_wallet[0].get("id"),
                    "status": model.status,
                    "label": model.label
                })

        db.bulk_update_mappings(CardSo, update_card)
        db.commit()
        return update_card

    def upload_wallets(self, db: Session):
        response_wallets = wallets.search(page_size=1000500, type="company").data.results
        query_wallets = db.query(WalletSo).filter(WalletSo.search_id is not None).all()

        list_wallets_id = []
        list_wallets = []
        for w in query_wallets:
            list_wallets_id.append(w.search_id)
            list_wallets.append({"search_id": w.search_id, "id": w.id})

        filter_wallet = list(filter(
            lambda w: w.id in list_wallets_id,
            response_wallets
        ))
        update_wallets = []
        for wallet in filter_wallet:
            response_wallet = list(filter(
                lambda w: w.get("search_id") == wallet.id ,
                list_wallets
            ))
            if response_wallet:
                update_wallets.append({
                    "id": response_wallet[0].get("id"),
                    "balance": wallet.available_amount,
                    "blocked_balance": wallet.blocked_amount
                })

        db.bulk_update_mappings(WalletSo, update_wallets)
        db.commit()
        return update_wallets

    def wallet_update_balance(self, db: Session, wallet_id: int):
        wallet = db.query(WalletSo).filter(WalletSo.id == wallet_id).first()
        response_data = wallets.get(wallet.search_id)
        wallet.balance = response_data.data.available_amount
        wallet.blocked_balance = response_data.data.blocked_amount
        self.save_obj(db, wallet)
        return wallet

    def get_wallets(self, **kwargs):
        response_data = wallets.search(**kwargs)
        print(response_data)
        return response_data

    def get_card(self, card_id: str = None, showSensitiveData: str = None, **kwargs):
        return card.get(card_id, showSensitiveData)

    def update_info_card(self, db: Session, id: int, showSensitiveData=True):
        card_obj = db.query(CardSo).filter(CardSo.id == id).first()
        response_data = card.get(card_obj.search_id, showSensitiveData)
        response_card = response_data.data
        card_obj.PAN = response_card.pan
        card_obj.cvv = response_card.cvv
        card_obj.type = response_card.type
        card_obj.created_on = response_card.creation_time
        card_obj.expiry = response_card.expiration_date.date()
        card_obj.PAN = response_card.pan
        card_obj.status = response_card.status
        self.save_obj(db, card_obj)
        return card_obj

    def create_card(self, db: Session, user_id: int,
                    name: str = None, emboss_line4: str = None, type="VIRTUAL", card_label=None):
        wallet = db.query(WalletSo).filter(WalletSo.user_id == user_id).first()
        user = wallet.user

        if not card_label:
            card_label = self.settings.name

        if not name:
            name = user.email

        response_data = card.create(
            owner_public_id=wallet.search_id,
            wallet_id=wallet.search_id,
            name=name,
            emboss_line4=emboss_line4,
            type=type,
            card_label=card_label)
        order = response_data.data

        if not order.is_valid or order.status != "PLACED":
            raise SoldoException("Error create_card", response_data.dict())
        result = {
            "wallet_id": wallet.id,
            "status": order.status,
            "category": "CARD"
        }
        self.__cache[order.id] = result
        return result

    def add_item_to_group(self, id: str, type="WALLET", groupId: str = None):
        if not groupId:
            groupId = self.settings.GROUP_ID
        return group.group_write(groupId, id, type)

    def get_order(self, order_id: str):
        return order.get(order_id)




def init_virtual_card(name:str, service: str,
                      uri: str,
                      celery_broker: str,
                      celery_backend: str,
                      user_model,
                      **config):
    from vc.manager import Soldo
    required_config = []
    network = None
    if "soldo" == service:
        network = Soldo

    return network(name, uri=uri, celery_broker=celery_broker, user_model=user_model, celery_backend=celery_backend,
                   **config)


from contextlib import contextmanager
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session, Session
from typing import Callable
from vc.db.base_class import Base


class BaseNetworkClient(object):
    """
    event_list - is list function for webhook
    _user - base model User
    """
    _user = None
    events = {}
    event_list = []

    def __init__(self, uri: str, pool_timeout=10,
                 pool_size=5, max_overflow=100,
                 connect_timeout=10,
                 session_name="virtual_card", user_model=None, celery_broker=None, celery_backend=None, **kwargs):
        self._user = user_model

        # db
        self.engine = create_engine(uri, pool_pre_ping=True,
                                    pool_timeout=pool_timeout,
                                    pool_size=pool_size, max_overflow=max_overflow,
                                    connect_args={'connect_timeout': connect_timeout,
                                                  "application_name": session_name})
        self.SessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=self.engine)

        # generate add
        for event in self.event_list:
            self.register_event(event)

    @contextmanager
    def get_session(self):
        session = None
        try:
            session = self.SessionLocal()
            yield session
        except Exception as exc:
            try:
                session.rollback()
            except Exception:
                pass
            raise exc
        finally:
            if session:
                session.close()

    def save_obj(self, db, obj):
        db.add(obj)
        db.commit()
        db.refresh(obj)
        return obj

    def oauth_authorize(self):
        raise NotImplementedError

    def whoami(self):
        raise NotImplementedError

    def user_update(self, db: Session, id: int, **kwargs):
        raise NotImplementedError

    def activate_service(self, db: Session, user_id: int, **kwargs):
        raise NotImplementedError

    def create_wallet(self, db: Session,id: int):
        return NotImplementedError

    def create_card(self, request_timestamp, owner_type, owner_public_id,
                    wallet_id, type, name, emboss_line4, card_label="aff"):
        return NotImplementedError

    def add_item_to_group(self, groupId: str, id: str, type="WALLET"):
        return NotImplementedError

    def run_event(self, event_name, **data):
        event = self.events.get(event_name)
        if event:
            return event(**data)

    def run_event_db(self, db: Session, event_name, **data):
        event = self.events.get(event_name)
        if event:
            return event(db, **data)

    def register_event(self, func: Callable, name: str = None):
        if isinstance(func, str):
            name = func
            func = getattr(self, func)
        else:
            if not name:
                name = func.__name__
        self.events[name] = func




import logging
from logging.handlers import TimedRotatingFileHandler


def set_config(logger, filename: str = None):
    formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    logger.setLevel(logging.DEBUG)
    # FileHandler

    fh = logging.StreamHandler()
    if filename:
        fh = TimedRotatingFileHandler(filename, when="W0", backupCount=2)
    fh.setFormatter(formatter)
    logger.addHandler(fh)



from .utils import fingerprintHash, fingerprintSignature
from typing import Optional
from pydantic import Extra, BaseModel


class JWTData(BaseModel):
    refresh_token: str
    token_type: str
    access_token: str
    expires_in: int


class HeadersSoldoBase(BaseModel):
    """
    headers for Standard Authentication
    http://apidoc-demo.soldo.com/v2/zgxiaxtcyapyoijojoef.html#standard-authentication
    """
    Authorization: Optional[str]

    def __init__(self, *args, **kwargs):
        from vc.manager import Soldo
        super().__init__(*args, **kwargs)
        self.Authorization = f"Bearer {Soldo.settings.ACCESS_TOKEN}"

    class Config:
        extra = Extra.allow


class HeadersSoldo(HeadersSoldoBase):
    """
    headers for Advanced Authentication
    http://apidoc-demo.soldo.com/v2/zgxiaxtcyapyoijojoef.html#advanced-authentication
    """
    Content_Type: str = "application/json"
    fingerprintH: Optional[str]
    fingerprintS: Optional[str]

    def __init__(self, data, fields=None, **kwargs):
        from vc.manager import Soldo
        super().__init__(**kwargs)
        if not fields:
            fields = data.keys()
        fingerprint = ""
        for field in fields:
            f = data.get(field)
            if f is not None :
                if isinstance(f, bool):
                    f = str(f).lower()
                else:
                    f = str(f)
                fingerprint += f
                print(fingerprint)
        fingerprint += Soldo.settings.TOKEN
        print(fingerprint)
        self.fingerprintH = fingerprintHash(fingerprint)
        self.fingerprintS = fingerprintSignature(self.fingerprintH)

    class Config:
        fields = {
            "fingerprintS": "X-Soldo-Fingerprint-Signature",
            "fingerprintH": "X-Soldo-Fingerprint",
        }
        extra = Extra.allow



from sqlalchemy.ext.declarative import as_declarative, declared_attr
from sqlalchemy import Column, Integer, String


class BaseMixer:
    id = Column(Integer, primary_key=True, index=True, autoincrement=True)
    __name__: str

    # Generate __tablename__ automatically
    @declared_attr
    def __tablename__(cls) -> str:
        return cls.__name__.lower()

    def __repr__(self, input_str=None):
        """
        Model representation on the admin panel
        """
        res_string = f"<{self.__tablename__}: "

        if input_str:
            res_string += f"{input_str}>"
            return res_string

        res_string += f"id: {self.id}; "

        try:
            name = self.name
            res_string += f"name: {name}; "
        except AttributeError:
            pass

        try:
            cake_id = self.cake_id
            res_string += f"cake_id: {cake_id}"
        except AttributeError:
            pass

        res_string += ">"

        return res_string


@as_declarative()
class Base(BaseMixer):
    pass



from setuptools import setup, find_packages

setup(
    name='virtual_cards',
    version='0.0.18',
    url=' ',
    license='MIT',
    author='Dmytro Bondar',
    author_email='admin@affcountry.com',
    description='A small example package Virtual Card',
    long_description='file: README.md',
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    platforms='any',
    install_requires=[
        'sqlalchemy',
        'requests',
        'pydantic'
    ],
    tests_require=[
        'sqlalchemy',
        'requests',
        'pydantic'
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3.7',
    ],
    test_suite='nose.collector'
)